/* bn2html.c - C source for converting Bengali files to html files *
 * Author: Abhijit Das (Barda) CSA IISc Bangalore India            *
 * Last modified : July 25 1997                                    */
#include <stdio.h>

/* Change the encoding file name if it is not in the current *
 * directory and/or not has the name bn.enc                  */
#define ENCFNAME "/usr/local/lib/bn2html/bn.enc"

FILE *fopen(), *INF, *OUTF, *ENCF;
int mode = 0, english = 0;

typedef struct {
   char code[11];
   int ascii;
} encEntry;
encEntry encTable[1000];
int nenc = 0;

void uc(char *ips) {
   int i, len = strlen(ips);
   for (i=0; i<len; i++) 
      if ((ips[i] >= 'a') && (ips[i] <= 'z')) ips[i] -= 32;
}

void addEntry(char code[], char ch[]) {
   int i, ascii;

   if (ch[0]=='\\') {
      ascii = 0;
      for (i=1; i<strlen(ch); i++) ascii = 10 * ascii + (((int)ch[i]) - 48);
   } else {
      ascii = (int)ch[0];
   }
   if ((ascii < 0) || (ascii > 255)) {
      fprintf(stderr,"Invalid encoding \"%s=%s\" ... Ignoring\n",code,ch);
      return;
   }
   strcpy(encTable[nenc].code,code);
   encTable[nenc].ascii = ascii;
   nenc++;
}

void sortEncTable() {
   encEntry temp;
   int i,j;

   for (i=nenc-2; i>=0; i--) {
      for (j=0; j<=i; j++) {
         if (strcmp(encTable[j].code,encTable[j+1].code) > 0) {
            temp = encTable[j+1];
            encTable[j+1] = encTable[j];
            encTable[j] = temp;
         }
      }
   }
}

void readEncoding() {
   char cmd[200], line[100], code[11], ch[5], *chp, c;
   int i, len;

   sprintf(cmd,"test -e %s",ENCFNAME);
   if (system(cmd)) {
      fprintf(stderr,"File containing bn-to-html encodings not found\n");
      fprintf(stderr,"Unable to proceed ...\n");
      exit(-1);
   } else {
      ENCF = fopen(ENCFNAME,"r");
      while (!feof(ENCF)) {
         line[0] = 0;
         do {
            fscanf(ENCF,"%c",&c);
            if ((c != '\n') && (c != '\r'))
               sprintf(line,"%s%c",line,c);
         } while ((!feof(ENCF)) && (c != '\n') && (c != '\r'));
         for (i=0; i<11; i++) code[i]=0;
         for (i=0; i<5; i++) ch[i]=0;
         len = strlen(line);
         if ((line[0] != '#') && (len > 0)) {
            i = 0; chp = line;
            while ((line[i] != '=') && (i < len)) {
               code[i] = line[i]; i++; chp++;
            }
            chp++;
            strcpy(ch,chp);
            if ((strlen(code) == 0) || (strlen(ch) == 0))
               fprintf(stderr,"Invalid encoding \"%s\" ... Ignoring\n",line);
            else {
               addEntry(code,ch);
            }
         }
      }
      fclose(ENCF);
      if (nenc > 1) sortEncTable();
   }
}

void outputChar(int index) {
   if ((index < 32) || (index > 126)) fprintf(OUTF,"&#%d;",index);
   else fprintf(OUTF,"%c",index);
}

int alphabetic (char c) {
   return (((c >= 'A') && (c <= 'Z')) || ((c >= 'a') && (c <= 'z')) || (c == '^'));
}

int vowel (char c) {
   return ((c=='a')||(c=='A')||
           (c=='e')||(c=='E')||
           (c=='i')||(c=='I')||
           (c=='o')||(c=='O')||
           (c=='u')||(c=='U')||
           (c=='R'));
}

int consonant (char c) {
   return (alphabetic(c) && (!(vowel(c))));
}

int digit(char c) {
   return((c >= '0') && (c <='9'));
}

int punct(char c) {
   return ((!alphabetic(c)) && (!digit(c)));
}

void errorMessage(char msg[]) {
   fprintf(stderr,"%s\n",msg);
   fprintf(stderr,"Usage: bn2html [-t <TYPE>] [-o <OUTPUTFILE>] <INPUTFILE>\n");
   fprintf(stderr,"Options:\n");
   fprintf(stderr,"-t <TYPE>         Output file type\n");
   fprintf(stderr,"                  <TYPE> = 0 or BengaliOnly (default)\n");
   fprintf(stderr,"                           1 or AllowEnglish\n");
   fprintf(stderr,"-o <OUTPUTFILE>   Output to file <OUTPUTFILE>\n");
   fprintf(stderr,"                  Default is STDOUT\n");
   exit(-1);
}

void processOptions(int argc, char *argv[]) {
   if (argc == 1) {
      errorMessage("Too few arguments");
   }
   if (argc > 6) {
      errorMessage("Too many arguments");
   }
   if ((argc % 2) == 1) {
      errorMessage("Improper number of arguments");
   }
   if (argc == 2) {
      INF  = fopen(argv[1],"r");
      if (feof(INF)) {
         fprintf(stderr,"Input file does not exist or is empty\n");
         exit(-1);
      }
      OUTF = stdout;
   }
   if (argc == 4) {
      INF  = fopen(argv[3],"r");
      if (feof(INF)) {
         fprintf(stderr,"Input file does not exist or is empty\n");
         exit(-1);
      }
      if (!strcmp(argv[1],"-t")) {
         if ((!strcmp(argv[2],"0")) || (!strcmp(argv[2],"BengaliOnly"))) {
            mode = 0;
            OUTF = stdout;
         } else if ((!strcmp(argv[2],"1")) || (!strcmp(argv[2],"AllowEnglish"))) {
            mode = 1;
            OUTF = stdout;
         } else errorMessage("Invalid argument to -t");
      } else if (!strcmp(argv[1],"-o")) {
         char cmd[200];
         sprintf(cmd,"test -e %s",argv[2]);
         if (system(cmd)) {
            OUTF = fopen(argv[2],"w");
         } else {
            char resp;
            fprintf(stderr,"Output file exists. Overwrite (y/n)? ");
            scanf("%c",&resp);
            if ((resp == 'y') || (resp == 'Y')) OUTF = fopen(argv[2],"w");
            else {
               fprintf(stderr,"Okay. Quitting ...\n");
               exit(-1);
            }
         }
      } else if (argv[1][0] == '-') {
         char errmsg[100];
         sprintf(errmsg,"Unknown option %s",argv[1]);
         errorMessage(errmsg);
      } else errorMessage("Too many arguments");
   }
   if (argc == 6) {
      INF  = fopen(argv[5],"r");
      if (feof(INF)) {
         fprintf(stderr,"Input file does not exist or is empty\n");
         exit(-1);
      }
      if (!strcmp(argv[1],"-t")) {
         if ((!strcmp(argv[2],"0")) || (!strcmp(argv[2],"BengaliOnly")))
            mode = 0;
         else if ((!strcmp(argv[2],"1")) || (!strcmp(argv[2],"AllowEnglish")))
            mode = 1;
         else errorMessage("Invalid argument to -t");
      } else if (!strcmp(argv[1],"-o")) {
         char cmd[200];
         sprintf(cmd,"test -e %s",argv[2]);
         if (system(cmd)) {
            OUTF = fopen(argv[2],"w");
         } else {
            char resp;
            fprintf(stderr,"Output file exists. Overwrite (y/n)? ");
            scanf("%c",&resp);
            if ((resp == 'y') || (resp == 'Y')) OUTF = fopen(argv[2],"w");
            else {
               printf("Okay. Quitting ...\n");
               exit(-1);
            }
         }
      } else if (argv[1][0] == '-') {
         char errmsg[100];
         sprintf(errmsg,"Unknown option %s",argv[1]);
         errorMessage(errmsg);
      } else errorMessage("Too many arguments");
      if (!strcmp(argv[3],"-t")) {
         if ((!strcmp(argv[4],"0")) || (!strcmp(argv[4],"BengaliOnly")))
            mode = 0;
         else if ((!strcmp(argv[4],"1")) || (!strcmp(argv[4],"AllowEnglish")))
            mode = 1;
         else errorMessage("Invalid argument to -t");
      } else if (!strcmp(argv[3],"-o")) {
         char cmd[200];
         sprintf(cmd,"test -e %s",argv[4]);
         if (system(cmd)) {
            OUTF = fopen(argv[4],"w");
         } else {
            char resp;
            fprintf(stderr,"Output file exists. Overwrite (y/n)? ");
            scanf("%c",&resp);
            if ((resp == 'y') || (resp == 'Y')) OUTF = fopen(argv[4],"w");
            else {
               printf("Okay. Quitting ...\n");
               exit(-1);
            }
         }
      } else if (argv[3][0] == '-') {
         char errmsg[100];
         sprintf(errmsg,"Unknown option %s",argv[3]);
         errorMessage(errmsg);
      } else errorMessage("Too many arguments");
   }
}

int getAscii(char code[]) {
   int min = 0, max = nenc - 1, mid, status;

   if (strcmp(code,encTable[min].code) < 0) return -1;
   if (strcmp(code,encTable[min].code) == 0) return encTable[min].ascii;
   if (strcmp(code,encTable[max].code) > 0) return -1;
   if (strcmp(code,encTable[max].code) == 0) return encTable[max].ascii;
   while (max > min + 1) {
      mid = (min + max) / 2;
      status = strcmp(code,encTable[mid].code);
      if (status == 0) return encTable[mid].ascii;
      if (status > 0) min = mid;
      else max = mid;
   }
   return -1;
}

void printVowel(char swar[]) {
   int ascii;

   ascii = getAscii(swar);
   if (ascii == 97) fprintf(OUTF,"a");
   else if (ascii == 65) fprintf(OUTF,"aA");
   else if ((ascii == 1) || (ascii == 105)) fprintf(OUTF,"&#1;");
   else if ((ascii == 2) || (ascii == 73)) fprintf(OUTF,"&#2;");
   else if ((ascii == 3) || (ascii == 117)) fprintf(OUTF,"&#3;");
   else if ((ascii == 4) || (ascii == 85)) fprintf(OUTF,"&#4;");
   else if ((ascii == 5) || (ascii == 87)) fprintf(OUTF,"&#5;");
   else if ((ascii == 6) || (ascii == 101)) fprintf(OUTF,"&#6;");
   else if ((ascii == 7) || (ascii == 69)) fprintf(OUTF,"&#7;");
   else if (ascii == 8) fprintf(OUTF,"&#8;");
   else if ((ascii == 111) || (ascii == 79)) fprintf(OUTF,"o");
   else if (ascii == 14) fprintf(OUTF,"a&#14;A");
   else fprintf(OUTF,"&nbsp;");
}

void printConsonant(char banjon[]) {
   int ascii;
   int ref = 0, rafala = 0, jafala = 0, chandra = 0;
   int done = 0, len = strlen(banjon);

   if (!strcmp(banjon,"^")) {
      fprintf(OUTF,"w");
      return;
   }

   do {
      done = 1;
      if (banjon[len - 1] == '^') {
         chandra = 1; banjon[len - 1] = 0; len--; done = 0;
      }
      if (banjon[len - 2] == '^') {
         if (banjon[len - 1] == 'r') {
            rafala = 1; banjon[len - 2] = banjon[len - 1] = 0;
            len -= 2; done = 0;
         } else if ((banjon[len - 1] == 'j') || (banjon[len - 1] == 'y')) {
            jafala = 1; banjon[len - 2] = banjon[len - 1] = 0;
            len -= 2; done = 0;
         }
      }
      if ((banjon[0] == '^') && (banjon[1] == 'r')) {
         int i;
         ref = 1; len -= 2; done = 0;
         if (len > 0) 
            for (i=0; i<=len; i++)
               banjon[i] = banjon[i+2];
      }
   } while (!done);
   
   ascii = getAscii(banjon);
   if (ascii == -1) fprintf(OUTF,"&nbsp;");
   else outputChar(ascii);

   if (chandra) fprintf(OUTF,"w");
   if (ref) fprintf(OUTF,"^");
   if (rafala) fprintf(OUTF,"&#15;");
   if (jafala) fprintf(OUTF,"&#14;");
}

void printCombination(char swar[], char banjon[]) {
   int ascii = getAscii(swar);

   if (ascii == 65) {
      printConsonant(banjon);
      fprintf(OUTF,"A");
   } else if ((ascii == 1) || (ascii == 105)) {
      fprintf(OUTF,"i");
      printConsonant(banjon);
   } else if ((ascii == 2) || (ascii == 73)) {
      printConsonant(banjon);
      fprintf(OUTF,"I");
   } else if ((ascii == 3) || (ascii == 117)) {
      printConsonant(banjon);
      fprintf(OUTF,"u");
   } else if ((ascii == 4) || (ascii == 85)) {
      printConsonant(banjon);
      fprintf(OUTF,"U");
   } else if ((ascii == 5) || (ascii == 87)) {
      printConsonant(banjon);
      fprintf(OUTF,"W");
   } else if ((ascii == 6) || (ascii == 101)) {
      fprintf(OUTF,"e");
      printConsonant(banjon);
   } else if ((ascii == 7) || (ascii == 69)) {
      fprintf(OUTF,"E");
      printConsonant(banjon);
   } else if (ascii == 8) {
      fprintf(OUTF,"e");
      printConsonant(banjon);
      fprintf(OUTF,"A");
   } else if ((ascii == 111) || (ascii == 79)) {
      fprintf(OUTF,"e");
      printConsonant(banjon);
      fprintf(OUTF,"O");
   } else if (ascii == 14) {
      printConsonant(banjon);
      fprintf(OUTF,"&#14;A");
   } else printConsonant(banjon);
}

void processInput() {
   char c, swar[200], banjon[200], tag[1000], TAG[1000];
   int done = 0;

   while (!feof(INF)) {
      swar[0] = banjon[0] = tag[0] = 0;
      fscanf(INF,"%c",&c);
      if (english) {
         if (c == '<') {
            int imbalance = 1;
            do {
               sprintf(tag,"%s%c",tag,c);
               fscanf(INF,"%c",&c);
               if (c == '<') imbalance++;
               else if (c == '>') imbalance--;
               if (imbalance == 0) sprintf(tag,"%s>",tag);
            } while ((!(feof(INF))) && (imbalance > 0));
            strcpy(TAG,tag); uc(TAG);
            if ((!strcmp(TAG,"<TT>")) || (!strcmp(TAG,"<KBD>")) ||
                (!strcmp(TAG,"<CODE>")) || (!strcmp(TAG,"<SAMP>")) ||
                (!strcmp(TAG,"<PRE>")) || (!strcmp(TAG,"<XMP>")) ||
                (!strcmp(TAG,"<LISTING>")) || (!strcmp(TAG,"<PLAINTEXT>"))) {
               english++;
            } else if (((!strcmp(TAG,"</TT>")) || (!strcmp(TAG,"</KBD>")) ||
                (!strcmp(TAG,"</CODE>")) || (!strcmp(TAG,"</SAMP>")) ||
                (!strcmp(TAG,"</PRE>")) || (!strcmp(TAG,"</XMP>")) ||
                (!strcmp(TAG,"</LISTING>")) || (!strcmp(TAG,"</PLAINTEXT>")))
                 && (english > 0)) {
               english--;
            }
            fprintf(OUTF,"%s",tag);
         } else {
            fprintf(OUTF,"%c",c);
         }
      } else if (digit(c)) fprintf(OUTF,"%c",c);
      else if (vowel(c)) {
         done = 0;
         do {
            sprintf(swar,"%s%c",swar,c);
            if (feof(INF)) done = 1;
            else {
               fscanf(INF,"%c",&c);
               if (!vowel(c)) {
                  fseek(INF,-1,SEEK_CUR);
                  done = 1;
               }
            }
         } while (!done);
         printVowel(swar);
      } else if (consonant(c)) {
         done = 0;
         do {
            sprintf(banjon,"%s%c",banjon,c);
            if (feof(INF)) done = 1;
            else {
               fscanf(INF,"%c",&c);
               if (!consonant(c)) {
                  done = 1;
               }
            }
         } while (!done);
         if (vowel(c)) {
            done = 0;
            do {
               sprintf(swar,"%s%c",swar,c);
               if (feof(INF)) done = 1;
               else {
                  fscanf(INF,"%c",&c);
                  if (!vowel(c)) {
                     fseek(INF,-1,SEEK_CUR);
                     done = 1;
                  }
               }
            } while (!done);
            printCombination(swar,banjon);
         } else {
            printConsonant(banjon);
            fseek(INF,-1,SEEK_CUR);
         }
      } else { /* punct(c) */
         if (c == '_') {
            if (!feof(INF)) {
               fscanf(INF,"%c",&c);
               if (c == '_')
                  fprintf(OUTF,"_");
               else
                  fseek(INF,-1,SEEK_CUR);
            }
         } else if (c == '-') {
            if (feof(INF)) fprintf(OUTF,"-");
            else {
               fscanf(INF,"%c",&c);
               if (c == '-') {
                  if (feof(INF)) fprintf(OUTF,"&#123;");
                  else {
                     fscanf(INF,"%c",&c);
                     if (c == '-') fprintf(OUTF,"|");
                     else {
                        fseek(INF,-1,SEEK_CUR);
                        fprintf(OUTF,"&#123;");
                     }
                  }
               } else {
                  fprintf(OUTF,"-");
                  fseek(INF,-1,SEEK_CUR);
               }
            }
         } else if (c == '.') {
            if (feof(INF)) fprintf(OUTF,".");
            else {
               fscanf(INF,"%c",&c);
               if (c == '.') fprintf(OUTF,"&#60;");
               else {
                  fprintf(OUTF,".");
                  fseek(INF,-1,SEEK_CUR);
               }
            }
         } else if (c == '<') {
            int imbalance = 1;
            do {
               sprintf(tag,"%s%c",tag,c);
               fscanf(INF,"%c",&c);
               if (c == '<') imbalance++;
               else if (c == '>') imbalance--;
               if (imbalance == 0) sprintf(tag,"%s>",tag);
            } while ((!(feof(INF))) && (imbalance > 0));
            strcpy(TAG,tag); uc(TAG);
            if (((!strcmp(TAG,"<TT>")) || (!strcmp(TAG,"<KBD>")) ||
                (!strcmp(TAG,"<CODE>")) || (!strcmp(TAG,"<SAMP>")) ||
                (!strcmp(TAG,"<PRE>")) || (!strcmp(TAG,"<XMP>")) ||
                (!strcmp(TAG,"<LISTING>")) || (!strcmp(TAG,"<PLAINTEXT>")))
                 && (mode == 1)) {
               english++;
            } else if (((!strcmp(TAG,"</TT>")) || (!strcmp(TAG,"</KBD>")) ||
                (!strcmp(TAG,"</CODE>")) || (!strcmp(TAG,"</SAMP>")) ||
                (!strcmp(TAG,"</PRE>")) || (!strcmp(TAG,"</XMP>")) ||
                (!strcmp(TAG,"</LISTING>")) || (!strcmp(TAG,"</PLAINTEXT>")))
                 && (english > 0)) {
               english--;
            }
            fprintf(OUTF,"%s",tag);
         } else fprintf(OUTF,"%c",c);
      }
   }
}

void conclude() {
   fprintf(OUTF,"\n");
   fclose(INF);
   fclose(OUTF);
}

int main(int argc, char *argv[]) {
   fprintf(stderr,"This is bn2html Version 1.0\n"); fflush(stderr);
   processOptions(argc,argv);
   fprintf(stderr,"Starting conversion ... \n"); fflush(stderr);
   readEncoding();
   processInput();
   conclude();
   fprintf(stderr,"... Done\n");
   exit(0);
}
